import os

APP_NAME = "aep"

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': os.getenv('POSTGRES_DB', 'test_generic_table'),
        'USER': os.getenv('POSTGRES_USER', 'postgres'),
        'PASSWORD': os.getenv('POSTGRES_PASSWORD', 'postgres'),
        'HOST': 'localhost',
        'PORT': '5432',
    }
}

INSTALLED_APPS = [
    'testapp',
    'django.contrib.contenttypes',
    'django.contrib.auth',
]

SECRET_KEY = 'fake-key'



from typing import TypeAlias, Tuple, Optional
from django.views import View
from django.db.models import QuerySet
from django.core.exceptions import ImproperlyConfigured
from pmd_django.generic_table.generic_table import view

FrontendFieldOptions: TypeAlias = dict
FrontendFieldType: TypeAlias = str
FrontendHeaderDisplayName: TypeAlias = str
BackendAttributeName: TypeAlias = str
FrontendHeaderSettings: TypeAlias = Tuple[FrontendHeaderDisplayName, Tuple[FrontendFieldType, FrontendFieldOptions]]
StageDisplayValue: TypeAlias = str
StageDBValue: TypeAlias = str


class GenericTableView(View):
    queryset = None
    model = None
    stage_field: str = ""
    counted_stages: dict[StageDisplayValue, StageDBValue] = {}
    should_add_all: bool = True
    table_headers: dict[BackendAttributeName, Optional[FrontendHeaderSettings]] = {}

    # https://github.com/django/django/blob/5a1cae3a5675c5733daf5949759476d65aa0e636/django/views/generic/list.py#L22C5-L48C1
    def get_queryset(self):
        if self.queryset is not None:
            queryset = self.queryset
            if isinstance(queryset, QuerySet):
                queryset = queryset.all()
        elif self.model is not None:
            queryset = self.model._default_manager.all()
        else:
            raise ImproperlyConfigured(
                "%(cls)s is missing a QuerySet. Define "
                "%(cls)s.model, %(cls)s.queryset, or override "
                "%(cls)s.get_queryset()." % {"cls": self.__class__.__name__}
            )
        return queryset

    def get(self, request, *args, **kwargs):
        return view(
            qs=self.get_queryset(),
            request=request,
            **self.generic_table_view_kwargs
        )

    @property
    def generic_table_view_kwargs(self):
        return {
            "field": self.stage_field,
            "values_list": list(self.table_headers.keys()),
            "counted_values": list(self.counted_stages.values()),
            "data_key": "data",
            "final_json_hook": self.final_json_hook,
        }

    def final_json_hook(self, data: dict) -> dict:
        counted_stages = self.counted_stages
        if self.should_add_all:
            counted_stages = {"All": "ALL"} | self.counted_stages
        data["displayStageAttr"] = self.stage_field
        data["displayStages"] = counted_stages
        data["displayHeaders"] = {k: v[0] for k, v in self.table_headers.items() if v}
        data["filterFieldTypes"] = {
            k: v[1][0] for k, v in self.table_headers.items() if v
        }
        return data



import json

from django.conf import settings
from django.http import HttpRequest, HttpResponse
import base64
from jwcrypto import jwk, jws
from jwcrypto.common import json_decode


def api_key_middleware(get_response):
    def _(request: HttpRequest):
        if request.path.endswith("health") or request.path.endswith("health/"):
            return get_response(request)

        if request.path.endswith("identity") or request.path.endswith("identity/"):
            signed_identity = request.COOKIES.get("signedIdentity")
            if not signed_identity:
                return HttpResponse("Missing signedIdentity", status=401)
            return get_response(request)

        if request.path.endswith(("logout", "logout/")):
            return get_response(request)

        if request.method == "OPTIONS":
            return HttpResponse("Good for preflight")

        signed_identity = request.COOKIES.get("signedIdentity")

        if signed_identity:
            try:
                public_jwk_b64 = getattr(settings, "AUTH_PUBLIC_KEY", None)
                if public_jwk_b64:
                    jwk_json = base64.b64decode(public_jwk_b64).decode("utf-8")
                    key = jwk.JWK.from_json(jwk_json)

                    jws_token = jws.JWS()
                    jws_token.deserialize(signed_identity)
                    jws_token.verify(key)

                    identity = json_decode(jws_token.payload)

                    request.identity = identity
                    for p in identity["permissions"]:
                        if p['resource'] in (settings.APP_NAME, 'all'):
                            return get_response(request)

            except Exception:
                res = HttpResponse("Not authenticated")
                res.status_code = 401
                return res

        res = HttpResponse("Not authenticated")
        res.status_code = 401
        return res

    return _

def generate_identity(payload: dict):
    """Generates a signed identity for the given payload: {"email": "dev@techserv.com", "permissions": [{"resource": "all", "role": "dev"}]}"""
    payload = json.dumps(payload)
    private_jwk = jwk.JWK.generate(kty="OKP", crv="Ed25519")
    public_jwk_b64 = base64.b64encode(private_jwk.export_public().encode("utf-8")).decode("utf-8")

    signed = jws.JWS(payload.encode("utf-8"))
    signed.add_signature(private_jwk, None, protected=json.dumps({"alg": "EdDSA"}))

    return public_jwk_b64, signed.serialize(compact=True)



import io
import json

from django.db import connection
from django.db.models import Count
from django.http import FileResponse, JsonResponse
from django.test import RequestFactory, TestCase
from openpyxl.reader.excel import load_workbook

from pmd_django.generic_table.generic_table import (
    _apply_filters,
    _apply_ordering,
    _apply_user_filters,
    _build_counts,
    _generate_download_response,
    _get_allowed_fields,
    _get_page_size,
    _paginate_queryset,
    view,
)
from testapp.models import TestRelatedModel, TestModel

class TestGenericTable(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.factory = RequestFactory()

        # Create the table at runtime without migrations
        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(TestModel)
            schema_editor.create_model(TestRelatedModel)

        tm = TestModel.objects.create(status="active")

        TestModel.objects.bulk_create(
            [
                TestModel(status="inactive"),
                TestModel(status="active"),
                TestModel(status="inactive"),
                TestModel(status="active"),
            ]
        )

        TestRelatedModel.objects.create(name="Related Item", related=tm)

    @classmethod
    def tearDownClass(cls):
        with connection.schema_editor() as schema_editor:
            schema_editor.delete_model(TestModel)
        super().tearDownClass()

    def test_build_counts(self):
        counted_values = ["active", "inactive", "pending"]
        queryset = TestModel.objects.all()
        result = _build_counts(queryset, "status", counted_values)

        self.assertEqual(result, {"active": 3, "inactive": 2, "pending": 0, "ALL": 5})

    def test_get_page_size_valid(self):
        request = self.factory.get("/?page_size=20")
        self.assertEqual(_get_page_size(request, None), 20)

    def test_get_page_size_invalid(self):
        request = self.factory.get("/?page_size=invalid")
        self.assertEqual(_get_page_size(request, None), 30)

    def test_apply_filters(self):
        request = self.factory.get("/?status=active")
        extra_filters = {}
        extra_filter_from_request = {"status": "status__iexact"}

        filtered_qs = _apply_filters(
            TestModel.objects.all(), request, extra_filters, extra_filter_from_request
        )
        self.assertTrue(filtered_qs.exists())

    def test_apply_ordering(self):
        request = self.factory.get("/?sort_by=id&sort_order=desc")
        ordered_qs = _apply_ordering(TestModel.objects.all(), request)
        self.assertEqual(ordered_qs.first().id, 5)

    def test_paginate_queryset(self):
        request = self.factory.get("/?page=2&page_size=2")
        queryset = TestModel.objects.all().order_by("id")

        page_obj, paginator = _paginate_queryset(queryset, request, 2)
        self.assertEqual(page_obj.number, 2)
        self.assertEqual(paginator.num_pages, 3)
        self.assertEqual(
            list(page_obj.object_list.values_list("id", flat=True)), [3, 4]
        )

    def test_view(self):
        request = self.factory.get("/?page=1&sort_by=id")
        queryset = TestModel.objects.all()

        response = view(queryset, request, "status", ["active", "inactive"])

        self.assertIsInstance(response, JsonResponse)
        response_data = json.loads(response.content)

        self.assertIn("pagination", response_data)
        self.assertIn("stageCounts", response_data)

    def test_valid_user_filters(self):
        request = self.factory.get(
            "/",
            data={
                "user_filters": json.dumps(
                    [
                        {"field": "status", "condition": "exact", "value": "active"},
                        {
                            "field": "id",
                            "condition": "gte",
                            "value": 3,
                            "logicalOperator": "AND",
                        },
                    ]
                )
            },
        )
        queryset = _apply_user_filters(TestModel.objects.all(), request)
        self.assertTrue(queryset.exists())

    def test_invalid_field_filtering(self):
        request = self.factory.get(
            "/",
            data={
                "user_filters": json.dumps(
                    [{"field": "password", "condition": "exact", "value": "test"}]
                )
            },
        )
        with self.assertRaises(ValueError) as e:
            _apply_user_filters(TestModel.objects.all(), request)
        self.assertEqual(str(e.exception), "Filtering on 'password' is not allowed")

    def test_invalid_condition_filtering(self):
        request = self.factory.get(
            "/",
            data={
                "user_filters": json.dumps(
                    [{"field": "status", "condition": "hacked", "value": "active"}]
                )
            },
        )
        with self.assertRaises(ValueError) as e:
            _apply_user_filters(TestModel.objects.all(), request)
        self.assertEqual(str(e.exception), "Condition 'hacked' is not supported")

    def test_json_parsing_error(self):
        request = self.factory.get("/", data={"user_filters": "[INVALID JSON"})
        with self.assertRaises(ValueError) as e:
            _apply_user_filters(TestModel.objects.all(), request)
        self.assertEqual(str(e.exception), "Invalid JSON in user_filters")

    def test_missing_value_filter(self):
        request = self.factory.get(
            "/",
            data={
                "user_filters": json.dumps(
                    [{"field": "status", "condition": "exact", "value": None}]
                )
            },
        )
        queryset = _apply_user_filters(TestModel.objects.all(), request)
        self.assertEqual(queryset.count(), 5)  # Should return full dataset

    def test_apply_user_filters_where_we_have_dangling_operator(self):
        request = self.factory.get(
            "/",
            data={
                "user_filters": json.dumps(
                    [
                        {
                            "field": "id",
                            "condition": "gte",
                            "value": 3,
                            "logicalOperator": "AND",
                        }
                    ]
                )
            },
        )

        filtered_qs = _apply_user_filters(TestModel.objects.all(), request)
        self.assertTrue(filtered_qs.exists())

    def test_sql_injection_attempt(self):
        request = self.factory.get(
            "/",
            data={
                "user_filters": json.dumps(
                    [
                        {
                            "field": "status",
                            "condition": "exact",
                            "value": "'; DROP TABLE users; --",
                        }
                    ]
                )
            },
        )
        queryset = _apply_user_filters(TestModel.objects.all(), request)
        self.assertEqual(queryset.count(), 0)  # Should not execute harmful queries

    def test_unmatched_operators_and_filters(self):
        request = self.factory.get(
            "/",
            data={
                "user_filters": json.dumps(
                    [
                        {"field": "status", "condition": "exact", "value": "active"},
                        {
                            "field": "id",
                            "condition": "gte",
                            "value": 3,
                            "logicalOperator": "AND",
                        },
                        {
                            "field": "status",
                            "condition": "icontains",
                            "value": "inactive",
                        },
                        # No logicalOperator for this one
                    ]
                )
            },
        )
        with self.assertRaises(ValueError) as e:
            _apply_user_filters(TestModel.objects.all(), request)
        self.assertIn(
            "Filter-operator mismatch: expected 2 operators, got 1", str(e.exception)
        )

    def test_standard_model_fields(self):
        qs = TestModel.objects.all()
        allowed_fields = _get_allowed_fields(qs)

        self.assertIn("id", allowed_fields)
        self.assertIn("status", allowed_fields)

    def test_related_model_fields(self):
        qs = TestModel.objects.all()
        allowed_fields = _get_allowed_fields(qs)

        self.assertIn("testrelatedmodel__name", allowed_fields)

    def test_annotated_fields(self):
        qs = TestModel.objects.annotate(status_count=Count("status"))
        allowed_fields = _get_allowed_fields(qs)

        self.assertIn("status_count", allowed_fields)

    def test_values_fields(self):
        qs = TestModel.objects.values(
            "id", "status"
        )  # Only these fields should be allowed
        allowed_fields = _get_allowed_fields(qs)

        self.assertIn("id", allowed_fields)
        self.assertIn("status", allowed_fields)
        self.assertNotIn(
            "created_at", allowed_fields
        )  # Not in `.values()`, so should be rejected

    def test_rejects_unauthorized_fields(self):
        qs = TestModel.objects.all()
        allowed_fields = _get_allowed_fields(qs)

        self.assertNotIn("password_hash", allowed_fields)  # Doesn't exist in the model
        self.assertNotIn("random_field", allowed_fields)  # Should not be allowed

    def test_apply_user_filters_with_annotated_field(self):
        qs = TestModel.objects.annotate(status_count=Count("status"))

        request = self.factory.get(
            "/",
            data={
                "user_filters": json.dumps(
                    [{"field": "status_count", "condition": "gte", "value": 1}]
                )
            },
        )
        filtered_qs = _apply_user_filters(qs, request)
        self.assertTrue(filtered_qs.exists())

    def test_apply_user_filters_with_related_field(self):
        request = self.factory.get(
            "/",
            data={
                "user_filters": json.dumps(
                    [
                        {
                            "field": "testrelatedmodel__name",
                            "condition": "exact",
                            "value": "Related Item",
                        }
                    ]
                )
            },
        )
        qs = _apply_user_filters(TestModel.objects.all(), request)
        self.assertTrue(qs.exists())

    def test_generate_excel_response(self):
        queryset = TestModel.objects.all()
        values_list = ["status"]

        response = _generate_download_response(queryset, values_list)

        self.assertIsInstance(response, FileResponse)
        self.assertEqual(
            response["Content-Type"],
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        )

        response_content = io.BytesIO()
        for chunk in response.streaming_content:
            response_content.write(chunk)
        response_content.seek(0)

        workbook = load_workbook(response_content)
        worksheet = workbook.active

        expected_headers = ["status"]
        headers = [cell.value for cell in worksheet[1]]
        self.assertEqual(headers, expected_headers)

        data_rows = list(worksheet.iter_rows(min_row=2, values_only=True))
        expected_data = [
            ("active",),
            ("inactive",),
            ("active",),
            ("inactive",),
            ("active",),
        ]
        self.assertEqual(data_rows, expected_data)



from django.conf import settings
from django.test import TestCase
from pmd_django.auth import generate_identity


class BaseAuthTestCase(TestCase):
    def setUp(self):
        identity = {
            "email": "dev@techserv.com",
            "permissions": [{"resource": "all", "role": "dev"}],
        }
        self.set_signed_identity(identity)

    def set_signed_identity(self, identity: dict):
        public_jwk_b64, signed_identity_token = generate_identity(identity)
        settings.AUTH_PUBLIC_KEY = public_jwk_b64
        self.auth_headers = {"cookie": f"signedIdentity={signed_identity_token}"}
        self.client.cookies.load(self.auth_headers["cookie"])



import json

from django.http import JsonResponse
from django.test import TestCase, RequestFactory, override_settings

from pmd_django.auth import api_key_middleware, generate_identity

identity = {"email": "dev@techserv.com", "permissions": [{"resource": "all", "role": "dev"}]}
(public_jwk_b64, signed_token) = generate_identity(identity)
payload = json.dumps(identity)

@override_settings(AUTH_PUBLIC_KEY=public_jwk_b64)
class TestAuthMiddleware(TestCase):
    def setUp(self):
        self.factory = RequestFactory()

        def dummy_view(request):
            return JsonResponse(request.identity if hasattr(request, "identity") else {})

        self.middleware = api_key_middleware(dummy_view)

    def test_valid_signed_permissions(self):
        request = self.factory.get("/")
        request.COOKIES["signedIdentity"] = signed_token

        response = self.middleware(request)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)

        self.assertEqual(data, identity)

    def test_invalid_signature(self):
        request = self.factory.get("/")
        request.COOKIES["signedIdentity"] = "bad.token.value"

        response = self.middleware(request)
        self.assertEqual(response.status_code, 401)

    def test_missing_cookies(self):
        request = self.factory.get("/")
        response = self.middleware(request)
        self.assertEqual(response.status_code, 401)

    def test_logout_bypasses_auth(self):
        request = self.factory.post("/logout")
        response = self.middleware(request)

        self.assertEqual(response.status_code, 200)


import io
import json
import operator

from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.db.models import Count, Q
from django.http import FileResponse, JsonResponse
from openpyxl.workbook import Workbook

COMMON_RESPONSE_DEFAULTS = {
    "ordering": "-id",
    "page_size": 30,
}


def _build_counts(qs, status_field, counted_values):
    counts = {key: 0 for key in counted_values}
    total = 0
    for row in qs.values(status_field).annotate(count=Count(status_field)):
        counts[row[status_field]] = row["count"]
        total += row["count"]
    counts["ALL"] = total
    return counts


def _get_page_size(request, page_size):
    try:
        return int(
            request.GET.get(
                "page_size", page_size or COMMON_RESPONSE_DEFAULTS["page_size"]
            )
        )
    except (ValueError, TypeError):
        return COMMON_RESPONSE_DEFAULTS["page_size"]


def _apply_filters(qs, request, extra_filters=None, extra_filter_from_request=None):
    """
    Applies static filters and additional filters derived from the request parameters.
    """
    if extra_filters:
        qs = qs.filter(extra_filters)
    if extra_filter_from_request:
        for param, lookup in extra_filter_from_request.items():
            value = request.GET.get(param)
            if value:
                qs = qs.filter(**{lookup: value})
    return qs


def _get_allowed_fields(qs):
    """
    Retrieves allowed fields for filtering, including related model fields and dynamically added fields.
    """
    model = qs.model

    # Get direct fields from the model
    allowed_fields = {f.name for f in model._meta.get_fields() if hasattr(f, "attname")}

    # ForeignKey, OneToOneField, etc.
    related_fields = {
        f.name for f in model._meta.get_fields() if f.is_relation and f.related_model
    }

    # Allow filtering on related fields using `related_model__field` notation
    for rel_field in related_fields:
        related_model = model._meta.get_field(rel_field).related_model
        if related_model:
            allowed_fields.update(
                {
                    f"{rel_field}__{f.name}"
                    for f in related_model._meta.get_fields()
                    if hasattr(f, "attname")
                }
            )

    # Detect dynamically added fields from QuerySet (e.g., .annotate() or .values())
    if hasattr(qs, "_fields") and qs._fields:  # .values() fields
        allowed_fields.update(qs._fields)

    if hasattr(qs, "query") and hasattr(qs.query, "annotations"):  # .annotate() fields
        allowed_fields.update(qs.query.annotations.keys())

    if hasattr(qs, "extra_allowed_fields"):
        allowed_fields.update(qs.extra_allowed_fields)

    return allowed_fields


def _apply_user_filters(qs, request):
    """
    Applies dynamic user-defined filters from the request.
    """
    user_filters = request.GET.get("user_filters")
    if not user_filters:
        return qs

    ALLOWED_FIELDS = _get_allowed_fields(qs)
    ALLOWED_CONDITIONS = {
        "exact",
        "iexact",
        "contains",
        "icontains",
        "gt",
        "gte",
        "lt",
        "lte",
        "startswith",
        "istartswith",
        "endswith",
        "iendswith",
        "range",
        "isnull",
        "in",
        "exclude",
    }

    try:
        filters = json.loads(user_filters)
    except json.JSONDecodeError:
        raise ValueError("Invalid JSON in user_filters")

    operators = []
    queries = []

    for index, f in enumerate(filters):
        field = f.get("field")
        condition = f.get("condition")
        value = f.get("value")

        # Skip invalid filters
        if not field or not condition or value in [None, ""]:
            continue

        if field not in ALLOWED_FIELDS:
            raise ValueError(f"Filtering on '{field}' is not allowed")

        if condition not in ALLOWED_CONDITIONS:
            raise ValueError(f"Condition '{condition}' is not supported")

        lookup_expression = f"{field}__{condition}"
        queries.append(Q(**{lookup_expression: value}))

        if index > 0 and f.get("logicalOperator"):
            operators.append(
                {
                    "AND": operator.and_,
                    "OR": operator.or_,
                    "AND NOT": lambda a, b: operator.and_(a, operator.inv(b)),
                    "OR NOT": lambda a, b: operator.or_(a, operator.inv(b)),
                }.get(f["logicalOperator"], operator.and_)  # Default to AND if invalid
            )

    # Return QuerySet early if no valid filters exist
    if not queries:
        return qs

    if len(operators) != len(queries) - 1:
        raise ValueError(
            f"Filter-operator mismatch: expected {len(queries) - 1} operators, got {len(operators)}"
        )

    accumulator = queries[0]
    for q, op in zip(queries[1:], operators, strict=True):  # Ensures exact match
        accumulator = op(accumulator, q)

    return qs.filter(accumulator)


def _apply_ordering(qs, request):
    """
    Applies sorting based on request parameters.
    """
    sort_by = request.GET.get("sort_by", "id")
    sort_order = request.GET.get("sort_order", "asc")
    ordering = f"-{sort_by}" if sort_order == "desc" else sort_by
    return qs.order_by(ordering or COMMON_RESPONSE_DEFAULTS.get("ordering", "id"))


def _paginate_queryset(qs, request, page_size):
    """
    Handles pagination logic.
    """
    paginator = Paginator(qs, page_size)
    page = request.GET.get("page", 1)
    try:
        page_obj = paginator.page(page)
    except PageNotAnInteger:
        page_obj = paginator.page(1)
    except EmptyPage:
        page_obj = paginator.page(paginator.num_pages)

    return page_obj, paginator


def apply_all_filters(
    qs,
    request,
    field=None,
    counted_values=None,
    extra_filters=None,
    extra_filter_from_request=None,
):
    """
    Applies all relevant filtering, including user filters, extra filters, and field-based filtering.
    """
    qs = _apply_user_filters(qs, request)

    qs = _apply_filters(qs, request, extra_filters, extra_filter_from_request)

    counts = None
    if field is not None and counted_values is not None:
        counts = _build_counts(qs, field, counted_values)

    if field is not None and (status_value := request.GET.get("filter")):
        qs = qs.filter(**{field: status_value})  # Domain logic filter

    return qs, counts


def paginate_queryset(qs, request, page_size=None):
    """
    Handles pagination separately.
    """
    page_size = _get_page_size(request, page_size)
    return _paginate_queryset(qs, request, page_size)


def _generate_download_response(qs, values_list, transform=None):
    wb = Workbook()
    worksheet = wb.active
    worksheet.title = "Exported Data"

    if transform:
        data = transform(qs)
    else:
        data = list(qs.values(*values_list)) if values_list else list(qs.values())

    if not data:
        worksheet.append(["No data"])
    else:
        headers = list(data[0].keys())
        worksheet.append(headers)
        for row in data:
            worksheet.append([row.get(field) for field in headers])

    output = io.BytesIO()
    wb.save(output)
    output.seek(0)

    return FileResponse(
        output,
        content_type="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        as_attachment=True,
        filename="export.xlsx",
    )


def view(
    qs,
    request,
    field=None,
    counted_values=None,
    extra_filters=None,
    extra_filter_from_request=None,
    values_list=None,
    data_key="data",
    page_size=None,
    transform=None,
    final_json_hook=None,
) -> JsonResponse | FileResponse:
    """
    Orchestrates filtering and pagination for a queryset based on request parameters.
    """
    # Add lookups that user can see to permitted filter fields
    qs.extra_allowed_fields = values_list if values_list is not None else []

    # Apply filters and get counts
    qs, counts = apply_all_filters(
        qs, request, field, counted_values, extra_filters, extra_filter_from_request
    )

    # Apply sorting
    qs = _apply_ordering(qs, request)

    # Handle download request
    accept_header = request.headers.get("Accept", "")
    is_excel = (
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        in accept_header
    )
    if is_excel:
        return _generate_download_response(qs, values_list, transform=transform)

    # Paginate the filtered and sorted queryset
    page_obj, paginator = paginate_queryset(qs, request, page_size)

    # Extract required data
    if transform:
        data = transform(page_obj)
    else:
        data = list(page_obj.object_list.values(*values_list)) if values_list else [
            {field.name: getattr(obj, field.name) for field in obj._meta.fields}
            for obj in page_obj.object_list
        ]

    return_data = {
        "stages": list(counts.keys()) if counts else [],
        "stageCounts": counts if counts else {},
        data_key: data,
        "pagination": {
            "current_page": page_obj.number,
            "total_pages": paginator.num_pages,
            "total_items": paginator.count,
            "has_next": page_obj.has_next(),
            "has_previous": page_obj.has_previous(),
        },
    }

    if final_json_hook:
        return_data = final_json_hook(return_data)

    # Construct response
    return JsonResponse(return_data)



from django.db import models


class TestModel(models.Model):
    status = models.CharField(max_length=100)

    class Meta:
        managed = False
        db_table = "test_model"
        app_label = 'testapp'  # Explicitly set the app_label

class TestRelatedModel(models.Model):
    name = models.TextField()
    related = models.ForeignKey(TestModel, on_delete=models.CASCADE)

    class Meta:
        managed = False
        db_table = "test_related_model"
        app_label = 'testapp'


import os
import subprocess
from concurrent import futures

from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    help = "Build the react frontend into static and then start the dev server"

    def handle(self, *args, **options):
        os.chdir("./frontend")

        def frontend():
            completed = None
            try:
                completed = subprocess.run("npm run dev", shell=True)
            except KeyboardInterrupt:
                pass
            if completed and completed.returncode != 0:
                raise CommandError("Unable to start dev server")

        def backend():
            completed = None
            try:
                completed = subprocess.run("python ../manage.py runserver", shell=True)
            except KeyboardInterrupt:
                pass
            if completed and completed.returncode != 0:
                raise CommandError("Unable to start the backend")

        try:
            with futures.ThreadPoolExecutor(max_workers=2) as executor:
                executor.submit(frontend)
                executor.submit(backend)
        except KeyboardInterrupt:
            pass
